import { Command } from 'commander';
import fs from 'fs';
import { version } from '../package.json';
import { stdinToString } from './binutils';
import { parseMultiLine, stringifyTokens } from './parser';
import { Config, ModifyMode, TokenType, VariableToken } from './types';
import { update } from './manipulation';

export function normalizeParamVar(name: string, value: string): string {
  let fixValue = value;
  if (value.match(/[\s"']/)) {
    fixValue = `"${value.replace(/\"/g, '\\"')}"`;
  }
  return `${name.trim()}=${fixValue}`;
}

export function normalizeParams(rawParams: string[], resilient = false): string[] {
  const paramsFixed: string[] = [];
  let buffer: string[] = [];

  const bufferFlush = () => {
    const joined = buffer.join('');
    const bPos = joined.indexOf('=');
    if (bPos > -1) {
      paramsFixed.push(normalizeParamVar(joined.substring(0, bPos), joined.substring(bPos + 1)));
    } else {
      if (!resilient) {
        throw new Error(`Invalid argument: ${joined}`);
      }
    }
    buffer = [];
  };

  rawParams.forEach((ps: string, idx) => {
    const eqPos = ps.indexOf('=');
    if ((eqPos >= 0 && ps.length > 1) || idx == rawParams.length - 1) {
      if (eqPos == -1) {
        buffer.push(ps);
      }
      if (buffer.length > 0) {
        bufferFlush();
      }
      if (eqPos > -1) {
        paramsFixed.push(normalizeParamVar(ps.substring(0, eqPos), ps.substring(eqPos + 1)));
      }
    } else {
      buffer.push(ps);
      if (buffer.length >= 2 && buffer[buffer.length - 2] == '=') {
        bufferFlush();
      }
    }
  });
  return paramsFixed;
}

function makeReadCommand() {
  const getCmd = new Command('get');
  getCmd
    .description('Returns given variable from env file (if specified)')
    .option('-f, --file <filePath>', 'Input file to parse (if not given, stdio is used)')
    .argument('<key>', 'env variable name a.k.a. key')
    .action((key, options, cmd) => {
      const globOpts = cmd.optsWithGlobals();
      const data = options.file ? fs.readFileSync(options.file).toString() : stdinToString();
      const sourceName = options.file ? `file ${options.file}` : `stdin`;
      try {
        const tokens = parseMultiLine(data);
        const found = tokens.filter((t) => t.token === TokenType.VARIABLE && (t as VariableToken).name == key);
        if (found.length > 0) {
          if (globOpts.outputFile) {
            fs.writeFileSync(globOpts.outputFile, found[0].value);
          } else {
            process.stdout.write(found[0].value);
            process.stdout.write('\n');
          }
          process.exit();
        } else {
          if (!globOpts.silent) process.stderr.write(`Variable ${key} not found in ${sourceName}\n`);
          process.exit(1);
        }
      } catch (e: any) {
        if (!globOpts.silent) process.stderr.write(`Parsing ${sourceName} failed\n`);
        if (!globOpts.silent) process.stderr.write(e.toString() + '\n');
        process.exit(1);
      }
    });

  return getCmd;
}

export function makeProgram() {
  const program = new Command();
  program
    .name('dotenv-tool')
    .description('Tool to read and update .env files')
    .version(version, '-v, --version')
    .addCommand(makeReadCommand())
    .argument('[paramsToSet...]', 'space separated list of additional envs to set, in format key=value', '')
    .option('-i, --files <filePaths...>', 'Input file(s)')
    .option('-o, --outputFile <filePath>', 'Output file')
    .option('-m, --modify', 'Modify first input file')
    .option('-b, --beautify', 'Beautifies resulting env file')
    .option('-u, --onlyUpdate', 'Only updates existing values, without appending new values')
    .option('-p, --smartAppend', 'Smart appends variables depending on their names')
    .option('-r, --resilient', 'Ignore files that cannot be read during update')
    .option('-s, --silent', 'Mute all messages and errors')
    .action((paramsToSet, options, cmd) => {
      try {
        const inputData: string[] = options.files
          ? options.files.map((fname: string, index: number) => {
              try {
                return fs.readFileSync(fname).toString();
              } catch (e: any) {
                if (!options.resilient) {
                  if (!options.silent) process.stderr.write(`Updating failed, cannot read file: ${fname}\n`);
                  if (!options.silent) process.stderr.write(e.toString() + '\n');
                  process.exit(1);
                } else {
                  return '';
                }
              }
            })
          : [stdinToString()];
        const sourceName = options.files ? `file ${options.files[0]}` : `stdin`;
        const config: Config = {};
        const remainingFiles = (options.files && options.files.length) > 1 ? options.files.slice(1) : [];
        let validSource = false;
        let sourceParsingError = '';

        config.beautify = options.beautify || false;
        config.modifyMode = options.onlyUpdate ? ModifyMode.NO_APPEND : options.smartAppend ? ModifyMode.SMART_APPEND : ModifyMode.APPEND;

        const baseData = inputData.shift() || '';

        let tokens = [];
        try {
          tokens.push(...parseMultiLine(baseData));
          validSource = true;
        } catch (e: any) {
          sourceParsingError = e;
          if (!options.resilient) {
            if (!options.silent) process.stderr.write(`Updating failed, cannot parse source ${sourceName}\n`);
            if (!options.silent) process.stderr.write(e.toString() + '\n');
            process.exit(1);
          }
        }
        let updates = [];

        while (inputData.length > 0) {
          const parsedFile = remainingFiles.shift();
          try {
            const updateTokens = parseMultiLine(inputData.shift() || '');
            updates.push(...updateTokens.filter((t) => t.token === TokenType.VARIABLE).map((t) => t as VariableToken));
          } catch (e: any) {
            if (!options.resilient) {
              if (!options.silent) process.stderr.write(`Updating failed, cannot parse file: ${parsedFile}\n`);
              if (!options.silent) process.stderr.write(e.toString() + '\n');
              process.exit(1);
            }
          }
        }

        if (paramsToSet && paramsToSet.length > 0) {
          try {
            const paramsFixed = normalizeParams(paramsToSet, options.resilient);
            paramsFixed.forEach((param) => {
              try {
                const updateTokens = parseMultiLine(param);
                updates.push(...updateTokens.filter((t) => t.token === TokenType.VARIABLE).map((t) => t as VariableToken));
              } catch (e: any) {
                if (!options.resilient) {
                  if (!options.silent) process.stderr.write(`Updating failed, invalid param\n`);
                  if (!options.silent) process.stderr.write(e.toString() + '\n');
                  process.exit(1);
                }
              }
            });
          } catch (e: any) {
            if (!options.resilient) {
              if (!options.silent) process.stderr.write(`Updating failed, cannot parse params passed\n`);
              if (!options.silent) process.stderr.write(e.toString() + '\n');
              process.exit(1);
            }
          }
        }

        tokens = update(tokens, updates, config);
        const outputStr = stringifyTokens(tokens, config);

        if (options.outputFile) {
          fs.writeFileSync(options.outputFile, outputStr);
        } else if (options.modify && options.files.length > 0) {
          if (validSource) {
            fs.writeFileSync(options.files[0], outputStr);
          } else {
            if (!options.silent) process.stderr.write(`Overwriting invalid source file not possible: ${options.files[0]}\n`);
            if (!options.silent) process.stderr.write(sourceParsingError.toString() + '\n');
            process.exit(1);
          }
        } else {
          process.stdout.write(outputStr);
        }
        process.exit();
      } catch (e: any) {
        if (!options.resilient) {
          if (!options.silent) process.stderr.write(`Updating failed - other error\n`);
          if (!options.silent) process.stderr.write(e.toString() + '\n');
        }
        process.exit(1);
      }
    });

  return program;
}
