import { parseMultiLine, stringifyTokens } from './parser';

describe('MultiLine format parser', () => {
  it('parse simple prop', () => {
    expect(parseMultiLine('VARNAME=value')).toEqual([vtoken('VARNAME', 'value')]);
    expect(parseMultiLine('VARNAME =value')).toEqual([vtoken('VARNAME', 'value', '', '', '', '', ' =')]);
    expect(parseMultiLine('VARNAME= value')).toEqual([vtoken('VARNAME', 'value', '', '', '', '', '= ')]);
    expect(parseMultiLine('  VARNAME=value')).toEqual([vtoken('VARNAME', 'value', '', '', '  ')]);
    expect(parseMultiLine('VARNAME=value \t ')).toEqual([vtoken('VARNAME', 'value', '', '', '', ' \t ')]);
    expect(parseMultiLine('VARNAME=A:\\DYNA')).toEqual([vtoken('VARNAME', 'A:\\DYNA')]);
    expect(parseMultiLine('VARNAME=quote\'is"ok')).toEqual([vtoken('VARNAME', 'quote\'is"ok')]);
    expect(parseMultiLine('VARNAME=value not escaped')).toEqual([vtoken('VARNAME', 'value not escaped')]);
    expect(parseMultiLine('VARNAME=    value not escaped')).toEqual([vtoken('VARNAME', 'value not escaped', '', '', '', '', '=    ')]);
    expect(parseMultiLine('VARNAME=value not escaped  ')).toEqual([vtoken('VARNAME', 'value not escaped', '', '', '', '  ')]);
    expect(parseMultiLine('VARNAME=this# is# value')).toEqual([vtoken('VARNAME', 'this# is# value')]);
    expect(parseMultiLine('VARNAME= this# is# value  # and comment')).toEqual([vtoken('VARNAME', 'this# is# value', '', '  # and comment', '', '', '= ')]);
  });

  it('prop have defined names', () => {
    expect(parseMultiLine('abc=value')).toEqual([vtoken('abc', 'value')]);
    expect(parseMultiLine('aBc12_ed=value')).toEqual([vtoken('aBc12_ed', 'value')]);
    expect(parseMultiLine('__=value')).toEqual([vtoken('__', 'value')]);
    expect(parseMultiLine('_FINE=value')).toEqual([vtoken('_FINE', 'value')]);
    expect(parseMultiLine('A=value')).toEqual([vtoken('A', 'value')]);
    expect(parseMultiLine('x00001=value')).toEqual([vtoken('x00001', 'value')]);
    expect(() => parseMultiLine('007=Bond')).toThrow(/PARSING FAILED/);
    expect(() => parseMultiLine('kłącze=value')).toThrow(/PARSING FAILED/);
    expect(() => parseMultiLine('var$name=value')).toThrow(/PARSING FAILED/);
    expect(() => parseMultiLine('not%alowed=value')).toThrow(/PARSING FAILED/);
    expect(parseMultiLine('#thiswontbevar=value')).toEqual([ctoken('#thiswontbevar=value')]);
    expect(() => parseMultiLine('and#thisisnoteven=comment')).toThrow(/PARSING FAILED/);
  });

  it('parse double quoted prop', () => {
    expect(parseMultiLine('VARNAME="value"')).toEqual([vtoken('VARNAME', 'value', '"')]);
    expect(parseMultiLine('VARNAME ="value"')).toEqual([vtoken('VARNAME', 'value', '"', '', '', '', ' =')]);
    expect(parseMultiLine('VARNAME= "value"')).toEqual([vtoken('VARNAME', 'value', '"', '', '', '', '= ')]);
    expect(parseMultiLine('VARNAME="value with spaces"')).toEqual([vtoken('VARNAME', 'value with spaces', '"')]);
    expect(parseMultiLine('VARNAME=" value with spaces"')).toEqual([vtoken('VARNAME', ' value with spaces', '"')]);
    expect(parseMultiLine('VARNAME="value with spaces "')).toEqual([vtoken('VARNAME', 'value with spaces ', '"')]);
    expect(() => parseMultiLine('VARNAME="not closed')).toThrow(/PARSING FAILED/);
  });

  it('escapes in double quot', () => {
    expect(parseMultiLine('VARNAME="test \\" it"')).toEqual([vtoken('VARNAME', 'test " it', '"')]);
    expect(parseMultiLine('VARNAME="\\"\\"\\"\\""')).toEqual([vtoken('VARNAME', '""""', '"')]);
    expect(parseMultiLine('VARNAME="C:\\\\WINDOWS"')).toEqual([vtoken('VARNAME', 'C:\\WINDOWS', '"')]);
    expect(parseMultiLine('VARNAME="\\\\\\\\\\\\"')).toEqual([vtoken('VARNAME', '\\\\\\', '"')]);
    expect(parseMultiLine('VARNAME="beginning\\nnext line"')).toEqual([vtoken('VARNAME', 'beginning\\nnext line', '"')]);
  });

  it('parse single quoted prop', () => {
    expect(parseMultiLine(`VARNAME='value'`)).toEqual([vtoken('VARNAME', 'value', "'")]);
    expect(parseMultiLine(`VARNAME ='value'`)).toEqual([vtoken('VARNAME', 'value', "'", '', '', '', ' =')]);
    expect(parseMultiLine(`VARNAME= 'value'`)).toEqual([vtoken('VARNAME', 'value', "'", '', '', '', '= ')]);
    expect(parseMultiLine(`VARNAME='value with spaces'`)).toEqual([vtoken('VARNAME', 'value with spaces', "'")]);
    expect(parseMultiLine(`VARNAME=' value with spaces'`)).toEqual([vtoken('VARNAME', ' value with spaces', "'")]);
    expect(parseMultiLine(`VARNAME='value with spaces '`)).toEqual([vtoken('VARNAME', 'value with spaces ', "'")]);
    expect(() => parseMultiLine(`VARNAME='not closed`)).toThrow(/PARSING FAILED/);
  });

  it('escapes in single quot', () => {
    expect(parseMultiLine(`VARNAME='test \\' it'`)).toEqual([vtoken('VARNAME', "test ' it", "'")]);
    expect(parseMultiLine(`VARNAME='\\'\\'\\'\\''`)).toEqual([vtoken('VARNAME', "''''", "'")]);
    expect(parseMultiLine(`VARNAME='C:\\WINDOWS'`)).toEqual([vtoken('VARNAME', 'C:\\WINDOWS', "'")]);
    expect(parseMultiLine(`VARNAME='\\\\\\\\\\\\'`)).toEqual([vtoken('VARNAME', '\\\\\\', "'")]);
    expect(parseMultiLine(`VARNAME='beginning\\nnext line'`)).toEqual([vtoken('VARNAME', 'beginning\\nnext line', "'")]);
  });

  it('supports inline comments', () => {
    expect(parseMultiLine('VARNAME=value #comment')).toEqual([vtoken('VARNAME', 'value', '', ' #comment')]);
    expect(parseMultiLine('VARNAME = value    # comment with extra ')).toEqual([vtoken('VARNAME', 'value', '', '    # comment with extra ', '', '', ' = ')]);
    expect(parseMultiLine('VARNAME="value" # fine')).toEqual([vtoken('VARNAME', 'value', '"', ' # fine')]);
    expect(parseMultiLine(`VARNAME='value' # -----`)).toEqual([vtoken('VARNAME', 'value', "'", ' # -----')]);
    expect(parseMultiLine('VARNAME=endswithhash#  # plus comment')).toEqual([vtoken('VARNAME', 'endswithhash#', '', '  # plus comment')]);
  });

  it('supports standalone comments', () => {
    expect(parseMultiLine('#')).toEqual([ctoken('#')]);
    expect(parseMultiLine('#  ')).toEqual([ctoken('#  ')]);
    expect(parseMultiLine('  #')).toEqual([ctoken('  #')]);
    expect(parseMultiLine(' # ')).toEqual([ctoken(' # ')]);
    expect(parseMultiLine('########')).toEqual([ctoken('########')]);
  });

  it('handles empty file', () => {
    expect(parseMultiLine('')).toEqual([]);
    expect(parseMultiLine('\n')).toEqual([nltoken()]);
    expect(parseMultiLine('\n\n\n\n\n')).toEqual([nltoken('\n\n\n\n\n')]);
    expect(parseMultiLine('    \n\n  \n')).toEqual([wstoken('    '), nltoken('\n\n'), wstoken('  '), nltoken()]);
  });

  it('parse complex expressions', () => {
    expect(
      parseMultiLine(
        `VAR1=ALA\nVAR2="MA KOTA"\nVAR3 = value3\n\nVaR4="some \\" escape   " \nVAR5=val5 #comment \n\n\nVAR6= abc \nVAR_COM=thisis#legit  #here comment\n# this is comment \n # only comment \n\n #next comment \n # last\n#no space\n#\n\n \nPUSTA=   \nempty2= "" \nVARS='ala'\nVARS2= 'ma kota'\nVARS3='o\\'clock'\nTESTSQ='not " need'\nTESTDQ="not ' need"`,
      ),
    ).toEqual([
      vtoken('VAR1', 'ALA'),
      nltoken(),
      vtoken('VAR2', 'MA KOTA', '"'),
      nltoken(),
      vtoken('VAR3', 'value3', '', '', '', '', ' = '),
      nltoken('\n\n'),
      vtoken('VaR4', 'some " escape   ', '"', '', '', ' '),
      nltoken(),
      vtoken('VAR5', 'val5', '', ' #comment '),
      nltoken('\n\n\n'),
      vtoken('VAR6', 'abc', '', '', '', ' ', '= '),
      nltoken(),
      vtoken('VAR_COM', 'thisis#legit', '', '  #here comment'),
      nltoken(),
      ctoken('# this is comment '),
      nltoken(),
      ctoken(' # only comment '),
      nltoken('\n\n'),
      ctoken(' #next comment '),
      nltoken(),
      ctoken(' # last'),
      nltoken(),
      ctoken('#no space'),
      nltoken(),
      ctoken('#'),
      nltoken('\n\n'),
      wstoken(' '),
      nltoken(),
      vtoken('PUSTA', '', '', '', '', '', '=   '),
      nltoken(),
      vtoken('empty2', '', '"', '', '', ' ', '= '),
      nltoken(),
      vtoken('VARS', 'ala', "'"),
      nltoken(),
      vtoken('VARS2', 'ma kota', "'", '', '', '', '= '),
      nltoken(),
      vtoken('VARS3', "o'clock", "'"),
      nltoken(),
      vtoken('TESTSQ', 'not " need', "'"),
      nltoken(),
      vtoken('TESTDQ', "not ' need", '"'),
    ]);
  });
});

describe('MultiLine format stringifier', () => {
  it('supports simple props, multiline', () => {
    expect(stringifyTokens([vtoken('VAR1', 'abcd1234')])).toEqual('VAR1=abcd1234');
    expect(stringifyTokens([vtoken('VAR1', 'Ala'), nltoken(), vtoken('VAR2', 'Ma'), nltoken(), vtoken('VAR3', 'Kota')])).toEqual(
      'VAR1=Ala\nVAR2=Ma\nVAR3=Kota',
    );
  });

  it('enforces escaping', () => {
    expect(stringifyTokens([vtoken('VAR1', 'Ala ma kota')])).toEqual('VAR1="Ala ma kota"');
    expect(stringifyTokens([vtoken('VAR1', ' value')])).toEqual('VAR1=" value"');
    expect(stringifyTokens([vtoken('VAR1', 'value ')])).toEqual('VAR1="value "');
    expect(stringifyTokens([vtoken('VAR1', 'with\ttab')])).toEqual('VAR1="with\\ttab"');
    expect(stringifyTokens([vtoken('VAR1', 'multi\nline')])).toEqual('VAR1="multi\\nline"');
    expect(stringifyTokens([vtoken('VAR1', 'other\rspecial\fchars')])).toEqual('VAR1="other\\rspecial\\fchars"');
  });

  it('is stable', () => {
    const source = `VAR1=ALA\nVAR2="MA KOTA"\nVAR3 = value3\n\nVaR4="some \\" escape   " \nVAR5=val5 #comment \n\n\nVAR6= abc \nVAR_COM=thisis#legit  #here comment\n# this is comment \n # only comment \n\n #next comment \n # last\n#no space\n#\n\n \nPUSTA=   \nempty2= "" \nVARS='ala'\nVARS2= 'ma kota'\nVARS3='o\\'clock'\nTESTSQ='not " need'\nTESTDQ="not ' need"`;
    expect(stringifyTokens(parseMultiLine(source))).toEqual(source);
  });

  it('fixes invalid fields', () => {
    expect(stringifyTokens([vtoken('VAR', 'value', '', 'without hash')])).toEqual('VAR=value # without hash');
    expect(stringifyTokens([vtoken('VAR', 'value', '', '# without space')])).toEqual('VAR=value # without space');
    expect(stringifyTokens([vtoken('VAR', 'value', '', 'junk # with junk')])).toEqual('VAR=value # with junk');
    expect(stringifyTokens([vtoken('VAR', 'value', '', '', '', '', ' zło ')])).toEqual('VAR=value');
    expect(stringifyTokens([vtoken('VAR', 'value', '', '', '', '', ' junk = junktoo')])).toEqual('VAR=value');
    expect(stringifyTokens([vtoken('VAR', 'value', 'a')])).toEqual('VAR=value');
    expect(stringifyTokens([vtoken('VAR', 'value', '', '', ' zzz ')])).toEqual('VAR=value');
    expect(stringifyTokens([vtoken('VAR', 'value', '', '', '', ' xxx ')])).toEqual('VAR=value');
  });

  it('quotes values when needed', () => {
    expect(stringifyTokens([vtoken('VAR', 'value with spaces', '')])).toEqual('VAR="value with spaces"');
    expect(stringifyTokens([vtoken('VAR', '"need quotes', '')])).toEqual('VAR="\\"need quotes"');
    expect(stringifyTokens([vtoken('VAR', "'need quotes", '')])).toEqual('VAR="\'need quotes"');
  });

  it('supports all tokens', () => {
    expect(stringifyTokens([nltoken('\n\n')])).toEqual('\n\n');
    expect(stringifyTokens([nltoken(' \njunk \r')])).toEqual('\n\r');
    expect(stringifyTokens([wstoken('   ')])).toEqual('   ');
    expect(stringifyTokens([wstoken()])).toEqual('');
    expect(stringifyTokens([wstoken('\t \tjunk\t \t')])).toEqual('\t \t\t \t');
    expect(stringifyTokens([ctoken('junk # comment')])).toEqual('# comment');
    expect(stringifyTokens([vtoken('VAR', 'value'), nltoken('\n\n'), ctoken('awesome'), nltoken(), wstoken('   '), nltoken()])).toEqual(
      'VAR=value\n\n# awesome\n   \n',
    );
  });
});
