import Parsimmon from 'parsimmon';
import { Token, Config, TokenType, VariableToken } from './types';
import { getDefaultConfig } from './utils';

const varName = Parsimmon.regexp(/[a-zA-Z_]+[a-zA-Z0-9_]*/).desc('environment variable name');
const equals = Parsimmon.string('=').desc('equals sign');
const normalSpaceOpt = Parsimmon.regexp(/[^\S\r\n]*/);
const doubleQuote = Parsimmon.string('"');
const singleQuote = Parsimmon.string("'");
const escapedChar = Parsimmon.regexp(/\\./);
const nonEscapedValuePartSQ = Parsimmon.regexp(/[^'\\]+/);
const escapedStrSQ = nonEscapedValuePartSQ.or(escapedChar.map((v: string) => v.replace("\\'", "'").replace('\\\\', '\\'))).many();
const nonEscapedValuePartDQ = Parsimmon.regexp(/[^"\\]+/);
const escapedStrDQ = nonEscapedValuePartDQ.or(escapedChar.map((v: string) => v.replace('\\"', '"').replace('\\\\', '\\'))).many();

const simpleValueNotQuoted = Parsimmon.regexp(/[^"'\s][^\s]*([^\S\r\n]+[^\s#][^\s]*)*/);

const valueWithSpacesDoubleQuoted = Parsimmon.seqObj<string, any>(['quote', doubleQuote], ['value', escapedStrDQ.map((v) => v.join(''))], doubleQuote);
const valueWithSpacesSingleQuoted = Parsimmon.seqObj<string, any>(['quote', singleQuote], ['value', escapedStrSQ.map((v) => v.join(''))], singleQuote);
const value = Parsimmon.alt(
  valueWithSpacesSingleQuoted,
  valueWithSpacesDoubleQuoted,
  simpleValueNotQuoted.map((v) => ({ quote: '', value: v })),
).desc('property value');
const commentOpt = Parsimmon.regexp(/[^\S\r\n]+#[^\r\n]*/);
const comment = Parsimmon.regexp(/[\n\r]*[^\S\r\n]*#[^\r\n]*/);
const newlineOpt = Parsimmon.regexp(/[\n\r]*/);

const envFile = Parsimmon.seqObj<string | any, any>(
  ['nl_in_front', newlineOpt.desc('new line')],
  [
    'content',
    Parsimmon.alt(
      comment.desc('comment').map((c) => ({ token: 'comment', value: c })),

      Parsimmon.seqObj<string | any, any>(
        ['beginning', normalSpaceOpt],
        ['name', varName],
        ['equals', Parsimmon.seq(normalSpaceOpt, equals, normalSpaceOpt).map((v) => v.join(''))],
        ['value', value.fallback('')],
        [
          'ending',
          Parsimmon.alt(
            commentOpt.desc('comment').map((v) => ({ type: 'comment', value: v })),
            normalSpaceOpt.desc('white space').map((s) => ({ type: 'whitespace', value: s })),
          ).desc('comment or whitespace'),
        ],
      ).map((o) => {
        o.quote = o.value.quote || '';
        o.value = o.value.value || '';
        o.comment = o.ending.type === 'comment' ? o.ending.value : '';
        o.ending = o.ending.type === 'whitespace' ? o.ending.value : '';
        o.token = 'variable';
        return o;
      }),
      normalSpaceOpt.map((w) => ({ token: 'whitespace', value: w })),
    ).desc('variable'),
  ],
).sepBy(Parsimmon.string('\n').desc('new line'));

export function parseMultiLine(input: string): Token[] {
  try {
    return envFile
      .tryParse(input)
      .map((o, i) => {
        if (i > 0) {
          o.nl_in_front = '\n' + o.nl_in_front;
        }
        return o;
      })
      .flatMap((o) => [{ token: 'newline', value: o.nl_in_front }, { ...o.content }])
      .filter((t) => !((t.token === 'newline' || t.token === 'whitespace') && t.value === ''));
  } catch (err: any) {
    throw err;
  }
}

export function stringifyTokens(tokens: Token[], config?: Config): string {
  const cfg = getDefaultConfig(config);
  return tokens
    .map((t) => {
      switch (t.token) {
        case TokenType.COMMENT:
          if (t.value.length > 0) {
            if (!t.value.match(/#/)) {
              t.value = '# ' + t.value;
            } else if (!t.value.match(/^[^\S\r\n]*#/)) {
              t.value = t.value.substring(t.value.indexOf('#'));
            }
          }
          return t.value;

        case TokenType.NEWLINE:
          return t.value.replace(/[^\n\r]/g, '');

        case TokenType.WHITESPACE:
          return t.value ? t.value.replace(/[^\s]/g, '') : '';

        case TokenType.VARIABLE:
          const vt = t as VariableToken;
          if (vt.equals.length > 0 && !vt.equals.match(/^[^\S\r\n]*=[^\S\r\n]*$/)) {
            vt.equals = '=';
          }
          if (vt.beginning.length > 0 && !vt.beginning.match(/^[^\S\r\n]*$/)) {
            vt.beginning = '';
          }
          if (vt.ending.length > 0 && !vt.ending.match(/^[^\S\r\n]*$/)) {
            vt.ending = '';
          }
          if (vt.quote.length > 0 && !vt.quote.match(/^["']$/)) {
            vt.quote = '';
          }
          if (vt.comment.length > 0) {
            if (!vt.comment.match(/#/)) {
              vt.comment = ' # ' + vt.comment;
            } else if (!vt.comment.match(/^[^\S\r\n]+#/)) {
              vt.comment = ' ' + vt.comment.substring(vt.comment.indexOf('#'));
            }
          }
          if (vt.quote === '') {
            if (vt.value.length > 0 && (vt.value.match(/^["'\s]/m) || vt.value.match(/\s/))) {
              vt.quote = '"';
            }
          }
          if (vt.value && vt.value.length > 0) {
            switch (vt.quote) {
              case '"':
                vt.value = vt.value
                  .replace('\\', '\\\\')
                  .replace('\n', '\\n')
                  .replace('\f', '\\f')
                  .replace('\t', '\\t')
                  .replace('\r', '\\r')
                  .replace('"', '\\"');
                break;
              case "'":
                vt.value = vt.value.replace('\\', '\\\\').replace("'", "\\'");
                break;
            }
          }
          return `${vt.beginning}${vt.name}${vt.equals}${vt.quote}${vt.value || ''}${vt.quote}${vt.comment}${vt.ending}`;
      }
    })
    .join('');
}
